/*
   Grant Plaster - ObjMeshLoaderSimple.cpp
   OBJ file loader that parses through an .obj file and creates vertex buffer
   objects to make an object mesh. Based off of code provided in CMeshLoaderSimple.cpp
   by Zoe Wood. Texturing not yet completely supported.
*/
#include "ObjMeshLoaderSimple.h"

#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <math.h>
#include <iomanip>
#include <string>
#include <streambuf>
#include "glm/glm.hpp"



class SColor
{

public:

    float Red, Green, Blue;

    SColor()
        : Red(0.7f), Green(0.7f), Blue(0.7f)
    {}

};

class SVector3
{

public:

	float X, Y, Z;

	SVector3()
		: X(0), Y(0), Z(0)
	{}

	SVector3(float in)
		: X(in), Y(in), Z(in)
	{}

	SVector3(float in_x, float in_y, float in_z)
		: X(in_x), Y(in_y), Z(in_z)
	{}

	SVector3 crossProduct(SVector3 const & v) const
	{
		return SVector3(Y*v.Z - v.Y*Z, v.X*Z - X*v.Z, X*v.Y - v.X*Y);
	}

	float dotProduct(SVector3 const & v) const
	{
		return X*v.X + Y*v.Y + Z*v.Z;
	}

	float length() const
	{
		return sqrtf(X*X + Y*Y + Z*Z);
	}

	SVector3 operator + (SVector3 const & v) const
	{
		return SVector3(X+v.X, Y+v.Y, Z+v.Z);
	}

	SVector3 & operator += (SVector3 const & v)
	{
		X += v.X;
		Y += v.Y;
		Z += v.Z;

		return * this;
	}

	SVector3 operator - (SVector3 const & v) const
	{
		return SVector3(X-v.X, Y-v.Y, Z-v.Z);
	}

	SVector3 & operator -= (SVector3 const & v)
	{
		X -= v.X;
		Y -= v.Y;
		Z -= v.Z;

		return * this;
	}

	SVector3 operator * (SVector3 const & v) const
	{
		return SVector3(X*v.X, Y*v.Y, Z*v.Z);
	}

	SVector3 & operator *= (SVector3 const & v)
	{
		X *= v.X;
		Y *= v.Y;
		Z *= v.Z;

		return * this;
	}

	SVector3 operator / (SVector3 const & v) const
	{
		return SVector3(X/v.X, Y/v.Y, Z/v.Z);
	}

	SVector3 & operator /= (SVector3 const & v)
	{
		X /= v.X;
		Y /= v.Y;
		Z /= v.Z;

		return * this;
	}

	SVector3 operator * (float const s) const
	{
		return SVector3(X*s, Y*s, Z*s);
	}

	SVector3 & operator *= (float const s)
	{
		X *= s;
		Y *= s;
		Z *= s;

		return * this;
	}

	SVector3 operator / (float const s) const
	{
		return SVector3(X/s, Y/s, Z/s);
	}

	SVector3 & operator /= (float const s)
	{
		X /= s;
		Y /= s;
		Z /= s;

		return * this;
	}

};

class SVertex
{

public:

    SVector3 Position;
    SVector3 Normal;
    SColor Color;
    SVector3 UV;

};

class OMesh
{

    friend class OVertexBufferObject;
    friend class OMeshLoader;

    struct STriangle
    {
        unsigned int VertexIndex1, VertexIndex2, VertexIndex3;
        unsigned int VertexUV1, VertexUV2, VertexUV3;
        unsigned int VertexNorm1, VertexNorm2, VertexNorm3;
        int currMat;
		SColor Color;
    };
    
    struct Material
    {
       std::string name;
       float alpha, ns, ni;
       float diff[3], amb[3], spec[3];
       int illum;
       int texture;
       //material();
    };

    std::vector<SVertex> Vertices;
    std::vector<SVertex> Normals;
    std::vector<STriangle> Triangles;
    std::vector<SVertex> UVs;
    std::vector<Material> Mats;

    OMesh();

public:

    ~OMesh();

	void centerMeshByAverage(SVector3 const & CenterLocation);
	void centerMeshByExtents(SVector3 const & CenterLocation);

	void resizeMesh(SVector3 const & Scale);

    void computeNormals();

};

bool OMeshLoader::loadVertexBufferObjectFromMesh(std::string const & fileName, int & TriangleCount, GLuint & PositionBufferHandle, GLuint & ColorBufferHandle, GLuint & NormalBufferHandle, GLuint & UVBufferHandle, GLuint & MatBufferHandle)
{
	OMesh * Mesh = loadASCIIMesh(fileName);
	if (! Mesh)
		return false;

	Mesh->resizeMesh(SVector3(1));
	Mesh->centerMeshByExtents(SVector3(0));
    Mesh->computeNormals();

	createVertexBufferObject(* Mesh, TriangleCount, PositionBufferHandle, ColorBufferHandle, NormalBufferHandle, UVBufferHandle, MatBufferHandle);
	
	return true;
}

OMesh * const OMeshLoader::loadASCIIMesh(std::string const & fileName)
{
	std::ifstream File;
	File.open(fileName.c_str());

	if (! File.is_open())
	{
		std::cerr << "Unable to open mesh file: " << fileName << std::endl;
		return 0;
	}

	OMesh * Mesh = new OMesh();
   int currentMat = 0;

	while (File)
	{
		std::string ReadString;
		std::getline(File, ReadString);

		std::stringstream Stream(ReadString);

		std::string Label;
		Stream >> Label;

		if (Label.find("#") != std::string::npos)
		{
			// Comment, skip
			continue;
		}

		if ("v" == Label)
		{
			SVector3 Position;
			Stream >> Position.X;
			Stream >> Position.Y;
			Stream >> Position.Z;

			SVertex Vertex;
			Vertex.Position = Position;
			Vertex.Normal = SVector3(0);

			Mesh->Vertices.push_back(Vertex);
		}
      else if ("vt" == Label) {
         SVector3 uv;
         
         Stream >> uv.X;
         Stream >> uv.Y;
         uv.Z = 0;
         uv.Y = 1 - uv.Y;
         
         SVertex Vertex;
         Vertex.UV = uv;
         
         Mesh->UVs.push_back(Vertex);
      }
      else if ("vn" == Label) {
         SVector3 Normal;
         Stream >> Normal.X;
         Stream >> Normal.Y;
         Stream >> Normal.Z;
         
         SVertex Vertex;
         Vertex.Normal = Normal;
         Vertex.Position = SVector3(0);
         Mesh->Vertices.push_back(Vertex);
      }
		else if ("f" == Label)
		{
			int Vertex1, Vertex2, Vertex3;
         int Texture1, Texture2, Texture3;
         int Normal1, Normal2, Normal3;
         char dummy;
			
         Stream >> Vertex1;
         if (Stream.peek() == '/') {
            Stream >> dummy;
            Stream >> Texture1;
            if (Stream.peek() == '/') {
               Stream >> dummy;
               Stream >> Normal1;
            }
         }
         
			Stream >> Vertex2;
         if (Stream.peek() == '/') {
            Stream >> dummy;
            Stream >> Texture2;
            if (Stream.peek() == '/') {
               Stream >> dummy;
               Stream >> Normal1;
            }
         }
         
			Stream >> Vertex3;
         if (Stream.peek() == '/') {
            Stream >> dummy;
            Stream >> Texture3;
            if (Stream.peek() == '/') {
               Stream >> dummy;
               Stream >> Normal1;
            }
         }

			OMesh::STriangle Triangle;
			Triangle.VertexIndex1 = Vertex1 - 1;
			Triangle.VertexIndex2 = Vertex2 - 1;
			Triangle.VertexIndex3 = Vertex3 - 1;
         Triangle.VertexUV1 = Texture1 - 1;
         Triangle.VertexUV2 = Texture2 - 1;
         Triangle.VertexUV3 = Texture3 - 1;
         Triangle.VertexNorm1 = Normal1 - 1;
         Triangle.VertexNorm2 = Normal2 - 1;
         Triangle.VertexNorm3 = Normal3 - 1;
         Triangle.currMat = currentMat;

			size_t Location;
			if ((Location = ReadString.find("{")) != std::string::npos) // there is a color
			{
				Location = ReadString.find("rgb=(");
				Location += 5; // rgb=( is 5 characters

				ReadString = ReadString.substr(Location);
				std::stringstream Stream(ReadString);
				float Color;
				Stream >> Color;
				Triangle.Color.Red = Color;
				Stream >> Color;
				Triangle.Color.Green = Color;
				Stream >> Color;
				Triangle.Color.Blue = Color;
			}

			Mesh->Triangles.push_back(Triangle);
		}
      else if ("usemtl" == Label)
      {
         std::string tmp;
         Stream >> tmp;
         
         for (int ndx = 0; ndx < Mesh->Mats.size(); ndx++) {
            if (Mesh->Mats[ndx].name.compare(tmp)== 0) {
               currentMat = ndx;
               break;
            }
         }
      }
      else if ("mtllib" == Label) {
         std::string newFile;
         Stream >> newFile;
         
         std::ifstream matFile;
	      matFile.open(newFile.c_str());

	      if (! matFile.is_open())
	      {
		      std::cerr << "Unable to open material file: " << newFile << std::endl;
		      return 0;
	      }
         
         while (matFile) {
            std::string matReadString;
		      std::getline(matFile, matReadString);

		      std::stringstream matStream(matReadString);

		      std::string matLabel;
		      matStream >> matLabel;
            
            std::string matName;
            float alfa, nS, nI;
            float d[3], a[3], s[3];
            int ill;
            int t = -1;
            char dummy;
            
            if (matLabel.find("#") != std::string::npos)
		      {
			      // Comment, skip
			      continue;
		      }
            
            if ("newmtl" == matLabel)
            {
               matStream >> matName;
            }
            else if ("Ns" == matLabel)
            {
               matStream >> nS;
            }
            else if ("Ka" == matLabel)
            {
               matStream >> a[0] >> a[1] >> a[2];
            }
            else if ("Kd" == matLabel)
            {
               matStream >> d[0] >> d[1] >> d[2];
            }
            else if ("Ks" == matLabel)
            {
               matStream >> s[0] >> s[1] >> s[2];
            }
            else if ("Ni" == matLabel)
            {
               matStream >> nI;
            }
            else if ("d" == matLabel)
            {
               matStream >> alfa;
            }
            else if ("illum" == matLabel)
            {
               matStream >> ill;
               
               OMesh::Material newMat;
            newMat.alpha = alfa;
            newMat.ns = nS;
            newMat.ni = nI;
            newMat.diff[0] = d[0];
            newMat.diff[1] = d[1];
            newMat.diff[2] = d[2];
            newMat.amb[0] = a[0];
            newMat.amb[1] = a[1];
            newMat.amb[2] = a[2];
            newMat.spec[0] = s[0];
            newMat.spec[1] = s[1];
            newMat.spec[2] = s[2];
            newMat.illum = ill;
            newMat.texture = t;
            
            Mesh->Mats.push_back(newMat);
            }
            else if ("map_Kd" == matLabel)
            {
               std::string textureFile;
               matStream >> textureFile;
               /* Not yet supported */
               //texture = loadTexture(textureFile);
            }
            
         }
      }
		else if ("" == Label)
		{
			// Just a new line, carry on...
		}
		else if ("l" == Label)
		{
			continue; //Skip - We're not doing any normal calculations... (oops!)
		}
		else
		{
			if ("s" != Label && "g" != Label) std::cerr << "While parsing ASCII mesh: Expected 'v' or 'vt' or 'vn' or 'f' label, found '" << Label << "'." << std::endl;
		}
	}

	if (! Mesh->Triangles.size() || ! Mesh->Vertices.size())
	{
		delete Mesh;
		return 0;
	}

	return Mesh;
}

void OMeshLoader::createVertexBufferObject(OMesh const & Mesh, int & TriangleCount, GLuint & PositionBufferHandle, GLuint & ColorBufferHandle, GLuint & NormalBufferHandle, GLuint & UVBufferHandle, GLuint & MatBufferHandle)
{
  	glGenBuffers(1, & PositionBufferHandle);
	glGenBuffers(1, & ColorBufferHandle);
	glGenBuffers(1, & NormalBufferHandle);
   glGenBuffers(1, & UVBufferHandle);
   glGenBuffers(1, & MatBufferHandle);

	std::vector<GLfloat> VertexPositions;
	std::vector<GLfloat> Colors;
	std::vector<GLfloat> lNormals;
   std::vector<GLfloat> lUVs;
   std::vector<GLfloat> lMats;

	for(unsigned int j = 0; j < Mesh.Triangles.size(); j++) 
	{
       	SVertex Vertex;
         
         lMats.push_back(Mesh.Triangles[j].currMat);

        Vertex = Mesh.Vertices[Mesh.Triangles[j].VertexIndex1];

        VertexPositions.push_back(Vertex.Position.X);
        VertexPositions.push_back(Vertex.Position.Y);
        VertexPositions.push_back(Vertex.Position.Z);
		VertexPositions.push_back(1.f);
		Colors.push_back(Mesh.Triangles[j].Color.Red);
		Colors.push_back(Mesh.Triangles[j].Color.Green);
		Colors.push_back(Mesh.Triangles[j].Color.Blue);
		lNormals.push_back(Vertex.Normal.X);
		lNormals.push_back(Vertex.Normal.Y);
		lNormals.push_back(Vertex.Normal.Z);
      lUVs.push_back(Vertex.UV.X);
      lUVs.push_back(Vertex.UV.Y);
      lUVs.push_back(Vertex.UV.Z);
      

		Vertex = Mesh.Vertices[Mesh.Triangles[j].VertexIndex2];

        VertexPositions.push_back(Vertex.Position.X);
        VertexPositions.push_back(Vertex.Position.Y);
        VertexPositions.push_back(Vertex.Position.Z);
		VertexPositions.push_back(1.f);
		Colors.push_back(Mesh.Triangles[j].Color.Red);
		Colors.push_back(Mesh.Triangles[j].Color.Green);
		Colors.push_back(Mesh.Triangles[j].Color.Blue);
		lNormals.push_back(Vertex.Normal.X);
		lNormals.push_back(Vertex.Normal.Y);
		lNormals.push_back(Vertex.Normal.Z);
      lUVs.push_back(Vertex.UV.X);
      lUVs.push_back(Vertex.UV.Y);
      lUVs.push_back(Vertex.UV.Z);
        
		Vertex = Mesh.Vertices[Mesh.Triangles[j].VertexIndex3];

        VertexPositions.push_back(Vertex.Position.X);
        VertexPositions.push_back(Vertex.Position.Y);
        VertexPositions.push_back(Vertex.Position.Z);
		VertexPositions.push_back(1.f);
		Colors.push_back(Mesh.Triangles[j].Color.Red);
		Colors.push_back(Mesh.Triangles[j].Color.Green);
		Colors.push_back(Mesh.Triangles[j].Color.Blue);
		lNormals.push_back(Vertex.Normal.X);
		lNormals.push_back(Vertex.Normal.Y);
		lNormals.push_back(Vertex.Normal.Z);
      lUVs.push_back(Vertex.UV.X);
      lUVs.push_back(Vertex.UV.Y);
      lUVs.push_back(Vertex.UV.Z);
        
	}

    TriangleCount = Mesh.Triangles.size();

	glBindBuffer(GL_ARRAY_BUFFER, PositionBufferHandle);
	glBufferData(GL_ARRAY_BUFFER, VertexPositions.size()*sizeof(GLfloat), & VertexPositions.front(), GL_STATIC_DRAW);

	glBindBuffer(GL_ARRAY_BUFFER, ColorBufferHandle);
	glBufferData(GL_ARRAY_BUFFER, Colors.size()*sizeof(GLfloat), & Colors.front(), GL_STATIC_DRAW);
	
    glBindBuffer(GL_ARRAY_BUFFER, NormalBufferHandle);
	glBufferData(GL_ARRAY_BUFFER, lNormals.size()*sizeof(GLfloat), & lNormals.front(), GL_STATIC_DRAW);
   
   glBindBuffer(GL_ARRAY_BUFFER, UVBufferHandle);
	glBufferData(GL_ARRAY_BUFFER, lUVs.size()*sizeof(GLfloat), & lUVs.front(), GL_STATIC_DRAW);

   glBindBuffer(GL_ARRAY_BUFFER, MatBufferHandle);
	glBufferData(GL_ARRAY_BUFFER, lMats.size()*sizeof(GLfloat), & lMats.front(), GL_STATIC_DRAW);

}

void OMesh::computeNormals() {

    SVector3 newNormal;
    SVector3 tmpNormal;
	
    for (std::vector<STriangle>::iterator it = Triangles.begin(); it != Triangles.end(); ++ it) {
    
    SVertex tmp = Vertices[it->VertexIndex1];
    SVertex tmp2 = Vertices[it->VertexIndex2];
    SVertex tmp3 = Vertices[it->VertexIndex3];
    
		newNormal = tmp2.Position - tmp.Position;
      tmpNormal = tmp3.Position - tmp.Position;
      newNormal = newNormal/(newNormal.length());
      tmpNormal = tmpNormal/(tmpNormal.length());
      
      newNormal = newNormal.crossProduct(tmpNormal);
      newNormal = newNormal/(newNormal.length());
      
      Vertices[it->VertexIndex1].Normal += newNormal;
      Vertices[it->VertexIndex2].Normal += newNormal;
      Vertices[it->VertexIndex3].Normal += newNormal;
    }
    for (std::vector<SVertex>::iterator it = Vertices.begin(); it != Vertices.end(); ++ it) {
    
       it->Normal = it->Normal/(it->Normal.length());
    }
}


OMesh::OMesh()
{}

OMesh::~OMesh()
{}

void OMesh::centerMeshByAverage(SVector3 const & CenterLocation)
{
	SVector3 VertexSum;
	for (std::vector<SVertex>::const_iterator it = Vertices.begin(); it != Vertices.end(); ++ it)
		VertexSum += it->Position;

	VertexSum /= (float) Vertices.size();
	SVector3 VertexOffset = CenterLocation - VertexSum;
	for (std::vector<SVertex>::iterator it = Vertices.begin(); it != Vertices.end(); ++ it)
		it->Position += VertexOffset;
}

void OMesh::centerMeshByExtents(SVector3 const & CenterLocation)
{
	if (Vertices.size() < 2)
		return;

	SVector3 Min, Max;
	{
		std::vector<SVertex>::const_iterator it = Vertices.begin();
		Min = it->Position;
		Max = it->Position;
		for (; it != Vertices.end(); ++ it)
		{
			if (Min.X > it->Position.X)
				Min.X = it->Position.X;
			if (Min.Y > it->Position.Y)
				Min.Y = it->Position.Y;
			if (Min.Z > it->Position.Z)
				Min.Z = it->Position.Z;

			if (Max.X < it->Position.X)
				Max.X = it->Position.X;
			if (Max.Y < it->Position.Y)
				Max.Y = it->Position.Y;
			if (Max.Z < it->Position.Z)
				Max.Z = it->Position.Z;
		}
	}

	SVector3 Center = (Max + Min) / 2;

	SVector3 VertexOffset = CenterLocation - Center;
	for (std::vector<SVertex>::iterator it = Vertices.begin(); it != Vertices.end(); ++ it)
		it->Position += VertexOffset;
}



void OMesh::resizeMesh(SVector3 const & Scale)
{
	if (Vertices.size() < 2)
		return;

	SVector3 Min, Max;
	{
		std::vector<SVertex>::const_iterator it = Vertices.begin();
		Min = it->Position;
		Max = it->Position;
		for (; it != Vertices.end(); ++ it)
		{
			if (Min.X > it->Position.X)
				Min.X = it->Position.X;
			if (Min.Y > it->Position.Y)
				Min.Y = it->Position.Y;
			if (Min.Z > it->Position.Z)
				Min.Z = it->Position.Z;

			if (Max.X < it->Position.X)
				Max.X = it->Position.X;
			if (Max.Y < it->Position.Y)
				Max.Y = it->Position.Y;
			if (Max.Z < it->Position.Z)
				Max.Z = it->Position.Z;
		}
	}

	SVector3 Extent = (Max - Min);
	SVector3 Resize = Scale / std::max(Extent.X, std::max(Extent.Y, Extent.Z));
	for (std::vector<SVertex>::iterator it = Vertices.begin(); it != Vertices.end(); ++ it)
		it->Position *= Resize;
}
