#include <cstdlib>
#include <cstdio>
#include <cmath>
#include <climits>

#include "GLSL_helper.h"

#ifdef __APPLE__
#include "GLUT/glut.h"
#include <OPENGL/gl.h>
#endif

#ifdef __unix__
#include <GL/glut.h>
#endif

#ifdef _WIN32
#pragma comment(lib, "glew32.lib")

#include <GL\glew.h>
#include <GL\glut.h>
#endif

#include "obj.h"

template <typename T>
T minus(T lhs, T rhs) {
    lhs.x -= rhs.x;
    lhs.y -= rhs.y;
    lhs.z -= rhs.z;
    return lhs;
}

template <typename T>
T plus(T lhs, T rhs) {
    lhs.x += rhs.x;
    lhs.y += rhs.y;
    lhs.z += rhs.z;
    return lhs;
}

template <typename T>
T normalize(T vec) {
    float len = sqrtf(vec.x*vec.x + vec.y*vec.y + vec.z*vec.z);
    if (len == 0) {
        vec.y = 1;
        return vec;
    }
    vec.x /= len;
    vec.y /= len;
    vec.z /= len;
    return vec;
}

template <typename T>
T cross(T lhs, T rhs) {
    T res;
    res.x = lhs.y*rhs.z - lhs.z*rhs.y;
    res.y = lhs.z*rhs.x - lhs.x*rhs.z;
    res.z = lhs.x*rhs.y - lhs.y*rhs.x;
    return res;
}

mesh_t load_mesh(const char *filename) {
    FILE *in = fopen(filename, "r");
    mesh_t mesh;

    vertex_t vertex;
    face_t face;
 
    vertex.normal.x = 0; 
    vertex.normal.y = 0; 
    vertex.normal.z = 0; 

    vector_t max = { -INFINITY, -INFINITY, -INFINITY };
    vector_t min = { INFINITY, INFINITY, INFINITY };
    
    while (fscanf(in, " v %f %f %f", &vertex.pos.x, &vertex.pos.y, &vertex.pos.z) == 3) {
        min.x = std::min(min.x, vertex.pos.x);
        min.y = std::min(min.y, vertex.pos.y);
        min.z = std::min(min.z, vertex.pos.z);

        max.x = std::max(max.x, vertex.pos.x);
        max.y = std::max(max.y, vertex.pos.y);
        max.z = std::max(max.z, vertex.pos.z);

        mesh.vertexes.push_back(vertex);
    }

    printf("Sizing: [(%f, %f, %f), (%f, %f, %f)]\n", min.x, min.y, min.z, max.x, max.y, max.z);

    for (std::vector<vertex_t>::iterator it = mesh.vertexes.begin();
         it < mesh.vertexes.end();
         it++)
    {
        it->pos.x = (it->pos.x - min.x) / (max.x - min.x) * 2 - 1;
        it->pos.y = (it->pos.y - min.y) / (max.y - min.y) * 2 - 1;
        it->pos.z = (it->pos.z - min.z) / (max.z - min.z) * 2 - 1;
    }

    while (fscanf(in, " f %u %u %u", &face.v1, &face.v2, &face.v3) == 3) {
        face.v1 -= 1;
        face.v2 -= 1;
        face.v3 -= 1;

        vector_t u = minus(mesh.vertexes[face.v1].pos, mesh.vertexes[face.v2].pos);
        vector_t v = minus(mesh.vertexes[face.v1].pos, mesh.vertexes[face.v3].pos);
        vector_t normal = cross(u, v);

        mesh.vertexes[face.v1].normal = plus(mesh.vertexes[face.v1].normal, normal);
        mesh.vertexes[face.v2].normal = plus(mesh.vertexes[face.v2].normal, normal);
        mesh.vertexes[face.v1].normal = plus(mesh.vertexes[face.v1].normal, normal);

        mesh.faces.push_back(face);
    }

    for (std::vector<vertex_t>::iterator it = mesh.vertexes.begin();
         it < mesh.vertexes.end();
         it++)
    {
        it->normal = normalize(it->normal);
    }

    return mesh;
}

void create_buffers(mesh_t const &mesh, int &numVertexes, int &numIndexes, GLuint &vertexes, GLuint &normals, GLuint &indexes) {
    glGenBuffers(1, &vertexes);
    glGenBuffers(1, &normals);
    glGenBuffers(1, &indexes);

    std::vector<GLfloat> vbuff;
    std::vector<GLfloat> nbuff;
    std::vector<GLushort>  ibuff;

    for (std::vector<vertex_t>::const_iterator it = mesh.vertexes.begin();
         it < mesh.vertexes.end();
         it++)
    {
        vbuff.push_back(it->pos.x);
        vbuff.push_back(it->pos.y);
        vbuff.push_back(it->pos.z);
        vbuff.push_back(1.0);

        nbuff.push_back(it->normal.x);
        nbuff.push_back(it->normal.y);
        nbuff.push_back(it->normal.z);
    }

    for (std::vector<face_t>::const_iterator it = mesh.faces.begin();
         it < mesh.faces.end();
         it++)
    {
        ibuff.push_back(it->v1);
        ibuff.push_back(it->v2);
        ibuff.push_back(it->v3);
    }

    numVertexes = vbuff.size() / 4;
    numIndexes  = ibuff.size();

    glBindBuffer(GL_ARRAY_BUFFER, vertexes);
    glBufferData(GL_ARRAY_BUFFER, vbuff.size()*sizeof(GLfloat), &vbuff.front(), GL_STATIC_DRAW);
    
    glBindBuffer(GL_ARRAY_BUFFER, normals);
    glBufferData(GL_ARRAY_BUFFER, nbuff.size()*sizeof(GLfloat), &nbuff.front(), GL_STATIC_DRAW);

    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, indexes);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, ibuff.size()*sizeof(GLushort), &ibuff.front(), GL_STATIC_DRAW);
}
